@@ -276,7 +276,7 @@ static char *extract_line(const char *stream, regoff_t begin, regoff_t end, int
 	for(e = end; stream[e] == '\t' || !iscntrl((unsigned char)stream[e]); e++);
 	length = e - *b;
 	if((retstr = (char *) malloc(length + 1))) {
-		sprintf(retstr, "%d:", *b);
+		snprintf(retstr, length + 1, "%d:", *b);
 		strncpy(retstr, &(stream[*b]), length);
 		retstr[length] = '\0';
 	}
@@ -612,12 +612,11 @@ int main(int argc, char **argv) {
 				options |= JG_PRINT_COUNT;
 				break;
 			case 'e':
-				if(!(regexpstr = (char *) malloc(strlen(optarg) + 1))) {
-					fprintf(stderr, "Malloc failure.\n");
-					fprintf(stderr, "Error: %s\n", strerror(errno));
-					exit(EXIT_FAILURE);
-				}
-				strcpy(regexpstr, optarg);
+                                regexpstr = strdup(optarg);
+                                if(NULL == regexpstr) {
+                                   perror("strdup");
+                                   exit(EXIT_FAILURE);
+                                }
 				break;
 			case 'i':
 				options |= JG_IGNORE_CASE;
