@@ -67,6 +67,7 @@
 #include "pushback.h"
 #include "compress.h"
 #include "shift.h"
+#include "argv.h"
 
 /* Some systems have mkdir that takes a single argument.  */
 #ifdef MKDIR_TAKES_ONE_ARG
@@ -107,7 +108,7 @@ static void version(void);
 static void add_entry(struct zipentry *);
 static void init_headers(void);
 
-static int consume(pb_file *, int);
+static void consume(pb_file *, size_t);
 static int list_jar(int, const char**, int);
 static int extract_jar(int, const char**, int);
 static int add_file_to_jar(int, int, const char*, struct stat*, int);
@@ -122,7 +123,6 @@ static int make_manifest(int, const char*, int);
 static int read_entries (int);
 static void init_args(const char **, int);
 static const char *get_next_arg (void);
-static char *jt_strdup (char*);
 static void expand_options (int *argcp, char ***argvp);
 static struct zipentry *find_entry (const char *);
 static int looks_like_dir (const char *);
@@ -161,7 +161,7 @@ static const char *progname;
 #define OPT_HELP     LONG_OPT (0)
 
 /* This holds all options.  */
-#define OPTION_STRING "-ctxuvVf:m:C:0Mi:E@"
+#define OPTION_STRING "-ctxuvVf:m:C:0Mi:J:E@"
 
 /* Define the MANIFEST content here to have it easier with calculations
    below.  This is for the case we create an empty MANIFEST.MF.  */
@@ -203,6 +203,7 @@ int main(int argc, char **argv)
   if(argc < 2)
     usage(argv[0]);
   
+  expandargv (&argc, &argv); /* expand @FILE parameters */
   new_argc = 0;
   new_argv = (const char **) malloc (argc * sizeof (char *));
 
@@ -251,7 +252,9 @@ int main(int argc, char **argv)
       action = ACTION_INDEX;
       jarfile = optarg;
       break;
-
+    case 'J':
+      /* ignored */
+      break;
     case OPT_HELP:
       help(argv[0]);
       break;
@@ -512,8 +515,8 @@ get_next_arg (void)
 	 names characters to read. */
 
       char *s;
-      int  pos = 0;
-      int len = 1024; /* Allocate more memory when we hit this
+      size_t  pos = 0;
+      size_t len = 1024; /* Allocate more memory when we hit this
 			    limit.  */
 
       s = (char *) malloc (len);
@@ -562,7 +565,7 @@ get_next_arg (void)
       if (pos)
 	{
 	  s [pos] = '\0';
-	  return jt_strdup (s);
+	  return strdup (s);
 	}
       else
 	return NULL;
@@ -837,10 +840,12 @@ int make_manifest(int jfd, const char *mf_name, int updating)
   }
   
   memset(ze, 0, sizeof(zipentry)); /* clear all the fields*/
-  ze->filename = (char*)malloc((nlen + 1) * sizeof(char) + 1);
-  strcpy(ze->filename, "META-INF/");
-  ze->filename[nlen] = '\0';
-    
+  ze->filename = strdup("META-INF/");
+  if (NULL == ze->filename) {
+     perror("strdup");
+     exit(EXIT_FAILURE);
+  }
+
   ze->offset = lseek(jfd, 0, SEEK_CUR);
   ze->mod_time = (ub2)(mod_time & 0x0000ffff);
   ze->mod_date = (ub2)((mod_time & 0xffff0000) >> 16);
@@ -860,7 +865,7 @@ int make_manifest(int jfd, const char *mf_name, int updating)
     if((mf = (char *) malloc(mf_len + 1))) {
     uLong crc;
 
-    sprintf(mf, "%s%s%s", MANIFEST_STR, VERSION, MANIFEST_END);
+    snprintf(mf, mf_len + 1, "%s%s%s", MANIFEST_STR, VERSION, MANIFEST_END);
 
     crc = crc32(0L, Z_NULL, 0);
     
@@ -887,9 +892,11 @@ int make_manifest(int jfd, const char *mf_name, int updating)
     }
     
     memset(ze, 0, sizeof(zipentry)); /* clear all the fields*/
-    ze->filename = (char*)malloc((nlen + 1) * sizeof(char) + 1);
-    strcpy(ze->filename, "META-INF/MANIFEST.MF");
-    ze->filename[nlen] = '\0';
+    ze->filename = strdup("META-INF/MANIFEST.MF");
+    if (NULL == ze->filename) {
+      perror("strdup");
+      exit(EXIT_FAILURE);
+    }
     
     ze->offset = lseek(jfd, 0, SEEK_CUR);
     ze->mod_time = (ub2)(mod_time & 0x0000ffff);
@@ -1037,14 +1044,15 @@ add_to_jar (int fd, const char *file, const int updating)
     
     nlen = strlen(file) + 256;
     fullname = (char*)malloc(nlen * sizeof(char));
-    memset(fullname, 0, (nlen * sizeof(char)));
-    
+
     if(fullname == NULL){
       fprintf(stderr, "Filename is NULL!\n");
       return 1;
     }
 
-    strcpy(fullname, file);
+    memset(fullname, 0, (nlen * sizeof(char)));
+    strncpy(fullname, file, nlen - 1);
+
     nlen = strlen(file);
 
     if(fullname[nlen - 1] != '/'){
@@ -1072,9 +1080,7 @@ add_to_jar (int fd, const char *file, const int updating)
     }
 
     memset(ze, 0, sizeof(zipentry)); /* clear all the fields*/
-    ze->filename = (char*)malloc((nlen + 1) * sizeof(char) + 1);
-    strcpy(ze->filename, fullname);
-    ze->filename[nlen] = '\0';
+    ze->filename = strdup(fullname);
     
     ze->offset = lseek(fd, 0, SEEK_CUR);
     ze->mod_time = (ub2)(mod_time & 0x0000ffff);
@@ -1218,8 +1224,7 @@ int add_file_to_jar(int jfd, int ffd, const char *fname, struct stat *statbuf,
   }
   
   memset(ze, 0, sizeof(zipentry)); /* clear all the fields*/
-  ze->filename = (char*)malloc((file_name_length + 1) * sizeof(char));
-  strcpy(ze->filename, fname);
+  ze->filename = strdup(fname);
 
   ze->mod_time = (ub2)(mod_time & 0x0000ffff);
   ze->mod_date = (ub2)((mod_time & 0xffff0000) >> 16);
@@ -1513,7 +1518,7 @@ int create_central_header(int fd){
 }
 
 int extract_jar(int fd, const char **files, int file_num){
-  int rdamt;
+  size_t rdamt;
   int out_a, in_a;
   ub4 signature;
   ub4 csize;
@@ -1523,7 +1528,7 @@ int extract_jar(int fd, const char **files, int file_num){
   ub2 flags;
   ub2 method;
   ub1 *filename = NULL;
-  int filename_len = 0;
+  ub2 filename_len = 0;
   ub4 rd_buff[RDSZ];
   pb_file pbf;
   ub1 scratch[16];
@@ -1533,8 +1538,16 @@ int extract_jar(int fd, const char **files, int file_num){
   int handle;
   int j;
 
-  init_inflation();
+  /*  Let's be able to hold 10 subdirs, initially. */
+  int subdir_list_size = 10;
+  /* If any file is contained within a directory in this list, 
+   * then it should be extracted. A file is considered to be contained
+   * in a directory if the directory is a substring of the file path. */
+  char **subdir_list = malloc(subdir_list_size * sizeof(char *));
+  /* Next available index to use. */
+  int subdir_list_index = 0;
 
+  init_inflation();
   pb_init(&pbf, fd);
 
   for(;;){
@@ -1559,7 +1572,7 @@ int extract_jar(int fd, const char **files, int file_num){
 #ifdef DEBUG    
       printf("skipping data descriptor\n");
 #endif
-      pb_read(&pbf, scratch, 12);
+      consume(&pbf, 12);
       continue;
     } else if(signature == 0x02014b50){
 #ifdef DEBUG    
@@ -1623,15 +1636,90 @@ int extract_jar(int fd, const char **files, int file_num){
 #ifdef DEBUG    
     printf("filename is %s\n", filename);
 #endif
-
-    if(file_num > 0){
+    /* If we have any subdirs that we're processing, then we must check */
+    if(file_num > 0 || subdir_list_index > 0){
       handle = FALSE;
       
-      for(j = 0; j < file_num; j++)
-        if(strcmp(files[j], (const char *)filename) == 0){
+      for(j = 0; j < subdir_list_index; j++) {
+#ifdef DEBUG
+        printf("extract_jar: %s substring of %s ? (subdir_list check)", subdir_list[j], filename);
+#endif
+        char * index = strstr((const char *)filename,(const char *)subdir_list[j]);
+        /* If the subdir is found as a substring at the beginning 
+         * of the current filename, then handle this filename. */
+        if(index != NULL && (((const char *)filename - index) == 0)) {
+#ifdef DEBUG
+          printf(" ... yes\n");
+#endif
           handle = TRUE;
           break;
+        } else {
+#ifdef DEBUG
+          printf("... no, index=%d, filename=%d\n", index, filename);
+#endif
         }
+      }
+
+      if (handle == FALSE) {
+        for(j = 0; j < file_num; j++) {
+#ifdef DEBUG
+          printf("extract_jar: %s same as %s ? ",files[j],filename);
+#endif
+          if(strcmp(files[j], (const char *)filename) == 0){
+#ifdef DEBUG
+            printf(" ... yes\n");
+#endif
+            handle = TRUE;
+            break;
+          } else {
+#ifdef DEBUG
+            printf(" ... no\n");
+#endif
+          }
+
+          /* Test to see if files[j] is a directory, too */
+          /* + 1 for extra '/' character */
+          int test_dir_len = strlen(files[j]) + 1;
+          char * test_dir = malloc(test_dir_len + 1 * sizeof(char));
+          strcpy(test_dir, files[j]);
+          test_dir[test_dir_len - 1] = '/';
+          test_dir[test_dir_len] = '\0';
+
+          char * index = strstr((const char *) filename, (const char *) test_dir);
+#ifdef DEBUG
+          printf("extract_jar: %s substring of %s ? ", test_dir, filename);
+#endif
+          /* If test_dir is indeed a directory, then add it to subdir_list, so 
+           * that we may extract any other files within this directory. */
+          if (index != NULL && (((const char *)filename - index) == 0)) {
+#ifdef DEBUG
+            printf(" ... yes\n");
+#endif
+            /* Add 10 more spots to subdir_list */
+            if (subdir_list_index == subdir_list_size) {
+              subdir_list_size = subdir_list_size + 10;
+              subdir_list = realloc(subdir_list, (subdir_list_size) * sizeof(char));
+
+              if (subdir_list == NULL) {
+                fprintf(stderr, "error realloc-ing subdir_list_size\n");
+                exit(EXIT_FAILURE);
+              }
+            }
+
+            subdir_list[subdir_list_index] = malloc((test_dir_len + 1)*sizeof(char));
+            strcpy(subdir_list[subdir_list_index], test_dir);
+            handle = TRUE;
+            subdir_list_index++;
+            break;
+          } else {
+#ifdef DEBUG
+            printf(" ... no \n");
+#endif
+          }
+
+          free(test_dir);
+        }
+      }
     }
 
     if(!handle)
@@ -1760,8 +1848,14 @@ int extract_jar(int fd, const char **files, int file_num){
         
         ze.crc = crc32(ze.crc, (Bytef*)rd_buff, rdamt);
 
-        if(f_fd >= 0)
-          write(f_fd, rd_buff, rdamt);
+        if(f_fd >= 0) {
+	  ssize_t num_written;
+          num_written = write(f_fd, rd_buff, rdamt);
+	  if (num_written == -1) {
+	    perror("write");
+	    exit(EXIT_FAILURE);
+	  }
+	}
 
         out_a += rdamt;
         in_a -= rdamt;
@@ -1792,8 +1886,8 @@ int extract_jar(int fd, const char **files, int file_num){
     }
 
     if(crc != ze.crc){
-      fprintf(stderr, "Error! CRCs do not match! Got %x, expected %x\n",
-              ze.crc, crc);
+      fprintf(stderr, "Error! CRCs do not match! Got %lx, expected %lx\n",
+              (unsigned long) ze.crc, (unsigned long) crc);
       exit(EXIT_FAILURE);
     }
 
@@ -1802,13 +1896,16 @@ int extract_jar(int fd, const char **files, int file_num){
     if(verbose && dir == FALSE && handle)
       printf("%10s: %s\n",
              (method == 8 ? "inflated" : "extracted"),
-             filename);
+             (const char *) filename);
   }
 
+  for (j = 0; j < subdir_list_index; j++) {
+    free(subdir_list[j]);
+  }
   return 0;
 }
 
-int list_jar(int fd, const char **files, int file_num){
+static int list_jar(int fd, const char **files, int file_num){
   ub4 signature;
   ub4 csize;
   ub4 usize;
@@ -1823,9 +1920,11 @@ int list_jar(int fd, const char **files, int file_num){
   ub1 *filename = NULL;
   ub1 scratch[16];
   ub1 cen_header[46];
-  int filename_len = 0;
+  ub2 filename_len = 0;
   off_t size;
-  int i, j;
+  off_t central_header_offset;
+  ub2 i;
+  int j;
   time_t tdate;
   struct tm *s_tm;
   char ascii_date[31];
@@ -1894,17 +1993,22 @@ int list_jar(int fd, const char **files, int file_num){
       tmp = L2BI(tmp);
 #endif
 
-      /*   printf("Central header offset = %d\n", tmp); */
+      central_header_offset = tmp;
 
-      if(lseek(fd, tmp, SEEK_SET) != (int)tmp){
+      /*   printf("Central header offset = %d\n", central_header_offset); */
+
+      if(lseek(fd, central_header_offset, SEEK_SET) != central_header_offset){
         perror("lseek");
         exit(EXIT_FAILURE);
       }
 
       /* Loop through the entries in the central header */
       for(i = 0; i < cen_size; i++){
+        ssize_t read_amt;
+
+        read_amt = read(fd, &cen_header, 46);
     
-        if(read(fd, &cen_header, 46) != 46){
+        if(read_amt == -1 || read_amt != 46){
           perror("read");
           exit(EXIT_FAILURE);
         }
@@ -1923,26 +2027,37 @@ int list_jar(int fd, const char **files, int file_num){
         /* If we're providing verbose output, we need to make an ASCII
          * formatted version of the date. */
         if(verbose){
+	  size_t time_string_length;
+
           mdate = UNPACK_UB4(cen_header, CEN_MODTIME);
           tdate = dos2unixtime(mdate);
           s_tm = localtime(&tdate);
-          strftime(ascii_date, 30, "%a %b %d %H:%M:%S %Z %Y", s_tm);
+          time_string_length = strftime(ascii_date, 30, "%a %b %d %H:%M:%S %Z %Y", s_tm);
+	  if (0 == time_string_length) {
+	    perror("strftime");
+	    exit(EXIT_FAILURE);
+	  }
           ascii_date[30] = '\0';
         }
 
         if(filename_len < fnlen + 1){
           if(filename != NULL)
             free(filename);
-      
+
           filename = malloc(sizeof(ub1) * (fnlen + 1));
+          if (NULL == filename) {
+            perror("malloc");
+            exit(EXIT_FAILURE);
+          }
           filename_len = fnlen + 1;
         }
     
-        if(read(fd, filename, fnlen) != fnlen){
+        read_amt = read(fd, filename, fnlen);
+        if(read_amt == -1 || read_amt != (ssize_t) fnlen){
           perror("read");
           exit(EXIT_FAILURE);
         }
-        filename[fnlen] = '\0';
+        filename[fnlen] = (ub1) '\0';
     
         /* if the user specified a list of files on the command line,
            we'll only display those, otherwise we'll display everything */
@@ -1950,14 +2065,14 @@ int list_jar(int fd, const char **files, int file_num){
           for(j = 0; j < file_num; j++)
             if(strcmp(files[j], (const char *)filename) == 0){
               if(verbose)
-                printf("%6d %s %s\n", usize, ascii_date, filename);
+                printf("%6lu %s %s\n", (unsigned long) usize, ascii_date, (const char *) filename);
               else
                 printf("%s\n", (const char *) filename);
               break;
             }
         } else {
           if(verbose)
-            printf("%6d %s %s\n", usize, ascii_date, filename);
+            printf("%6lu %s %s\n", (unsigned long) usize, ascii_date, (const char *) filename);
           else
             printf("%s\n", (const char *) filename);
         }            
@@ -2063,7 +2178,7 @@ int list_jar(int fd, const char **files, int file_num){
       }
       
       pb_read(&pbf, filename, fnlen);
-      filename[fnlen] = '\0';
+      filename[fnlen] = (ub1) '\0';
       
       /* the header is at the end.  In a JAR file, this means that the data
          happens to be compressed.  We have no choice but to inflate the
@@ -2100,14 +2215,14 @@ int list_jar(int fd, const char **files, int file_num){
         for(j = 0; j < file_num; j++)
           if(strcmp(files[j], (const char *)filename) == 0){
             if(verbose)
-              printf("%6d %s %s\n", usize, ascii_date, filename);
+              printf("%6lu %s %s\n", (unsigned long) usize, ascii_date, (const char *) filename);
             else
               printf("%s\n", (const char *) filename);
             break;
           }
       } else {
         if(verbose)
-          printf("%6d %s %s\n", usize, ascii_date, filename);
+          printf("%6lu %s %s\n", (unsigned long) usize, ascii_date, (const char *) filename);
         else
           printf("%s\n", (const char *) filename);
       }        
@@ -2116,21 +2231,38 @@ int list_jar(int fd, const char **files, int file_num){
   return 0;
 }
 
-int consume(pb_file *pbf, int amt){
-  int tc = 0; /* total amount consumed */
-  ub1 buff[RDSZ];
-  int rdamt;
+/**
+ * Consume bytes from a pushback buffer.
+ *
+ * @param pbf The pushback buffer to consume from.
+ * @param amt The amount of bytes to consume.
+ */
+static void consume(pb_file *pbf, size_t amt){
+  size_t tc = 0; /* total amount consumed */
+  ub1 *buff;
+  size_t rdamt;
 
 #ifdef DEBUG
   printf("Consuming %d bytes\n", amt);
 #endif
+  
+  buff = (ub1 *) malloc(RDSZ * sizeof(ub1));
+  if (NULL == buff) {
+    perror("malloc");
+    exit(EXIT_FAILURE);
+  }
 
   if (seekable){
-    if (amt <= (int)pbf->buff_amt)
-      pb_read(pbf, buff, amt);
+    if (amt <= pbf->buff_amt)
+      tc += pb_read(pbf, buff, amt);
     else {
-      lseek(pbf->fd, amt - pbf->buff_amt, SEEK_CUR);
-      pb_read(pbf, buff, pbf->buff_amt); /* clear pbf */
+      off_t location;
+      location = lseek(pbf->fd, amt - pbf->buff_amt, SEEK_CUR);
+      if (location == -1) {
+        perror("lseek");
+        exit(EXIT_FAILURE);
+      }
+      tc += pb_read(pbf, buff, pbf->buff_amt); /* clear pbf */
     }
   } else
   while(tc < amt){
@@ -2141,11 +2273,11 @@ int consume(pb_file *pbf, int amt){
     tc += rdamt;
   }
 
-#ifdef DEBUG
-  printf("%d bytes consumed\n", amt);
-#endif
+  free(buff);
 
-  return 0;
+#ifdef DEBUG
+  printf("%d bytes consumed\n", tc);
+#endif
 }
 
 void usage(const char *filename){
@@ -2192,6 +2324,10 @@ Store many files together in a single `jar' file.\n\
                   and its Class-Path (currently unimplemented)\n\
   -v              generate verbose output on standard output\n\
   -V, --version   display version information\n\
+");
+  printf("\n\
+  -J*             -J options are ignored\n\
+  @FILE           expand options and files from FILE\n\
 ");
   printf("\n\
 If any file is a directory then it is processed recursively.\n\
@@ -2208,16 +2344,6 @@ Example 2: use an existing manifest file 'mymanifest' and archive all the\n\
   exit(EXIT_SUCCESS);
 }
 
-static char *
-jt_strdup(char *s)
-{
-  char *result = (char*)malloc(strlen(s) + 1);
-  if (result == (char*)0)
-    return (char*)0;
-  strcpy(result, s);
-  return result;
-}
-
 /* Convert "tar-style" first argument to a form expected by getopt.
    This idea and the code comes from GNU tar.  This can allocate a new
    argument vector.  This might leak some memory, but we don't care.  */
@@ -2233,8 +2359,8 @@ expand_options (int *argcp, char ***argvp)
     {
       char buf[3];
       char **new_argv;
-      int new_argc;
-      int args_to_expand;
+      size_t new_argc;
+      size_t args_to_expand;
       char *p;
       char **in, **out;
 
@@ -2258,7 +2384,7 @@ expand_options (int *argcp, char ***argvp)
 	{
 	  char *opt;
 	  buf[1] = *p;
-	  *out++ = jt_strdup (buf);
+	  *out++ = strdup (buf);
 	  /* If the option takes an argument, move the next argument
 	     to just after this option.  */
 	  opt = strchr (OPTION_STRING, *p);
@@ -2280,7 +2406,7 @@ expand_options (int *argcp, char ***argvp)
       while (in < argv + argc)
 	*out++ = *in++;
 
-      *argcp = new_argc;
+      *argcp = (int) new_argc;
       *argvp = new_argv;
     }
 }
@@ -2375,8 +2501,7 @@ int add_array_to_jar(int jfd, char* content, size_t content_length, const char *
   }
   
   memset(ze, 0, sizeof(zipentry)); /* clear all the fields*/
-  ze->filename = (char*)malloc((file_name_length + 1) * sizeof(char));
-  strcpy(ze->filename, fname);
+  ze->filename = strdup(fname);
 
   ze->mod_time = (ub2)(mod_time & 0x0000ffff);
   ze->mod_date = (ub2)((mod_time & 0xffff0000) >> 16);
@@ -2408,10 +2533,18 @@ int add_array_to_jar(int jfd, char* content, size_t content_length, const char *
      as before */
   
   /* Write the local header */
-  write(jfd, file_header, 30);
-    
+  written = write(jfd, file_header, 30);
+  if ((written == -1) || ((size_t) written != 30)){
+    perror("write");
+    exit(EXIT_FAILURE);
+  }
+
   /* write the file name to the zip file */
-  write(jfd, fname, file_name_length);
+  written = write(jfd, fname, file_name_length);
+  if ((written == -1) || ((size_t) written != file_name_length)){
+    perror("write");
+    exit(EXIT_FAILURE);
+  }
 
 
   if(verbose){
@@ -2445,7 +2578,7 @@ int add_array_to_jar(int jfd, char* content, size_t content_length, const char *
   written = write(jfd, content, content_length);
   if ((written == -1) || ((size_t) written != content_length)){
     perror("write");
-    return 0;    	
+    exit(EXIT_FAILURE);
   }
       
   /* write out data descriptor */
@@ -2544,7 +2677,7 @@ int build_index(int jfd)
 
   index_content_size = strlen("JarIndex-Version: 1.0\n\n\n") + strlen(short_jar_fname) + 1;
   index_content = malloc(index_content_size);
-  sprintf(index_content, "JarIndex-Version: 1.0\n\n%s\n", short_jar_fname);
+  snprintf(index_content, index_content_size, "JarIndex-Version: 1.0\n\n%s\n", short_jar_fname);
 	
   for (ze = ziptail; ze; ze = ze->next_entry)
     {
